( 3iCore.fs - my core functionality for GForth - 2021-08-08---Sun---09:56 - removed from 3iForth.fs 
Assad Ebrahim
)

decimal

\ special PC keys:
27 constant esc-key
: pause  ( -- ) key drop  ;  \ press any key to continue  Note: to loop until key is pressed use e.g. begin ... key? until key drop 
\ Design: loopflag for eg begin ... loopflag until
0 constant continueloop
1 constant exitloop

\ for logic/bitwise operations
: not  ( x/0 -- 0/1 ) if 0 else 1 then ;  \ Boolean logic
\ bitwise complement
: not4 ( $xx -- $yy ) invert $F and ;  \ ex: $8 (1000b) not4 = 0111b = 7
: not8 ( $xx -- $yy ) invert $FF and ; \ ex: $8 (00001000b) not8 = 11110111b = 247

\ * MEMORY  - these words are useful for working efficiently with memory buffers, i.e. arrays of memory * )
1 constant byte 
2 constant wword
4 constant dword
: bytes  byte * ;
: wwords wword * ;   
: dwords dword * ;   \ GForth is 32-bits
\ Efficient fetch & store from/into and an array
\ BYTES
:  c@+ ( adr -- adr' val )  
    dup 1 bytes + swap    c@ ;
:  c!+ ( val adr -- adr' )  
    dup 1 bytes + -rot    c! ;

: go-to-end  ( array narray sz -- array-end narray ) 
    swap >r r@ * + 1 - r> ;
:  c@- ( adr -- adr' val )  
    dup 1 bytes - swap    c@ ;
\ WORDS (2 bytes)
\ ADD w, to complement c,
: bytes2word ( hi8 lo8 -- val16 )
	swap    8 lshift   or ;
: lohi ( val16 -- hi8 lo8  )  \ convert val16 to little-endian byte stream ( low first on top of stack) ex: $FFFE lohi ( -- $FF $FE )
    dup   $FF00 and   8 rshift    swap $FF and ;
    \ 440 f2hl 1 184
: ww@ ( adr -- val16 )  \ little endian retrieval
	c@+    swap c@    swap  bytes2word ;
: ww@+  ( adr -- adr' val16 )  
	c@+    swap c@+   swap -rot  swap  bytes2word ;
: ww!  ( val16 adr -- )  \ little endian storage
	swap lohi rot  c!+ c! ;
: ww!+ ( val16 adr -- adr' )  
	swap lohi rot  c!+ c!+ ;
\ DWORDS (4 bytes)
: @+ 
    dup 1 dwords + swap @ ;
: !+
    dup 1 dwords + -rot ! ;

\ STACK words
: 3drop 2drop drop ;
: 4drop 2drop 2drop ;  \ for 4-bit port work
: ndrop ( n-elements N -- ) 0 do drop loop ;  \ needed in 3iShift.fs beats sequencer
: ndup  ( n-elements N --  n-elements  n-elements N ) dup 0 do dup pick swap loop ; \ needed in 3iShift.fs beats sequencer
: nrev  ( a1 a2 a3 .. aN  N -- aN aN-1 .. a3 a2 a1 N )  \ needed in various places for user interface to accept l2r input
                    \ e.g. 3iSonglist.fs music playback, 3iFont.fs, 3iI2C.fs, 3iArt.fs
   dup 1 > if
      >r r@ 1 do i roll loop r> 
   else 
   then ;
: ..s  ( <list> n -- <list> n ) dup 0 do dup roll dup . swap loop ;   \ shows full stack, not just first 8
: showlist ..s ;

\ Words to work with moving lists from stack to memory (8-bit character) and showing them from memory
\ ex:  create tmpmem 8 bytes allot   1 2 3 4 5 6 7 8 8 tmpmem >mem   tmpmem .mem   tmpmem mem>
: >mem ( <list> nlen ptr --        nlen ) swap >r r@ dup 0 do dup i - 1 + roll rot c!+ swap loop 2drop r> ;   \ list on stack to memory
: mem> (        nlen ptr -- <list> nlen ) swap >r r@     0 do c@+ swap loop drop r> ; 
: .mem (        nlen ptr --        nlen ) swap >r r@     0 do c@+ . loop drop r> ;

   
\ math
: /up   ( a b -- ceil[a/b]  )   /mod swap 0 > if 1 + then ;  \ integer divide & round up     
: /dn   ( a b -- floor[a/b] )   / ;

\ characters
: between? ( n a b -- t/f ) rot dup rot > -rot > or invert ;  \ not ( n>b or n<a )
: ascii  255 0 do i 
    dup 14 = if 11 emit space 12 emit space then
    dup 4 mod 0 = if 2 spaces dup 8 mod 0 = if cr then then 
    emit space loop 
    cr ." 7 = bell" 
    cr ." 8 = ?"
    cr ." 9 = space"
    cr ." 10 = newline"
    cr ." 13 = newline"
;

\ TODO CODE REVIEW: CREATE FUNCTION FOR CHANGE OF BINARY BASE

\ BINARY TOOLS
( * BINARY viewing bitfield of a number  - determines number of bits from T )
: T4   $F  ;  \ ex: 4-bits 1111b  Determines whether calculations are 4-, 8-, 16-bits, or 32-bits
: T8   $FF  ;  \ ex: 8-bits 11111111b  Determines whether calculations are 4-, 8-, 16-bits, or 32-bits
: T16   $FFFF ; \ ex: 16-bits 
: T32   $FFFFFFFF ; \ ex: 32 bits  T = -1

\ ***** CHOOSE WORD SIZE IN THE BELOW DEFINITION OF ~
:  ~    ( p -- ~p ) T8 xor ;          \ negation op takes bitfield size from T
:  T~    ( T p -- ~p ) xor ;          \ negation op takes bitfield size from T
\ : init-mask  ( T -- ) 2 / ~  ;  \ initial mask is $8=1000b, $80=1000000b, or $8000 depending on T
: init-mask  ( T -- ) dup 2 / T~  ;  \ initial mask is $8=1000b, $80=1000000b, or $8000 depending on T

\ ***** CHOOSE WORD SIZE IN THE BELOW DEFINITION OF ~
: bit  ( n -- 1/0 ) 0> if 1 else 0 then ;  
: .bit ( 1/0 -- emits '1'/'0' ) '0' + emit ;
: bits  ( number  --  )  \  loop through bits from MSB mask and print
    T8 init-mask begin 
        2dup and bit .bit
        1 rshift  \ advance mask
        dup 0=    \ test if done
    until 2drop ." b" ;
: .b ( bf -- bf ) dup bits 2 spaces ;
: .sh ( N -- ) ndup nrev 0 do hex. loop ;  \ show stack in hex

\ **** make a word from bits
: >b8 ( 8-bits -- byte ) 0 8 1 do or swap i lshift loop or ;  \ test: 219 .b  1 1 0 1  1 0 1 1  >b8 .
\ : b8>
: >b4 ( 4-bits -- nibble ) 0 4 1 do or swap i lshift loop or ;  \ test: 11  .b  1 0 1 1  >b8 .
: b8>  ( number  --  )  \  loop through bits from MSB mask and leave on stack
    T8 init-mask begin 
        2dup and bit >r
        1 rshift  \ advance mask
        dup 0=    \ test if done
    until 2drop 
    r> r> r> r> r> r> r> r> 8 nrev drop ;


: help-mem
cr ." help-mem"
cr ." .s   ..s   -  dump list "
cr ." .mem       -  dump memory"
cr ." >mem       -  list to mem"
cr ." mem>       -  mem to list"
cr
;

help-mem
