( 3iI2C.fs - AE - 2021-06-03---Thu---14:24 )
( GForth file for communicating directly with OLED display's SSD1306 driver IC, allowing for bare metal development, 
  i.e. without use of Adafruit's SSD1306 device driver software or their GFX graphics library. )
\ publication mention on Hackaday :)
\ https://hackaday.com/2021/07/19/an-oled-photo-frame-powered-by-the-attiny85/

\ TODO: Embed fixed-width Font into OLED driver and embedded font display for fast draw

\ Connections: 2-pins for I2C comms on Arduino Nano: SDA=A4 (18, green/purple), SCL=A5 (19, orange), plus power VCC +5VDC and Gnd

\ References:
\ https://github.com/datacute/Tiny4kOLED

\ Design: Screen hardware description
\ Design: The SSD1306 display driver IC is built to control 128 wide x 64 high pixels
\ Design: There are 128 columns (SEG) and 64 rows (COM)
\ Design: A character is an 8x8 pixel block.
\ Design: The SSD1306 is built for character facilities
\ Design: A "page" is a block of 8 rows, i.e. the location for 8-high characters.
\ Design: So a 64-high display has 8 pages (0-7).  A 32-high display has 4 pages (0-3).
\ Design: The SSD1306 remembers position (cursor) and auto-advances depending on the mode address mode.
\ Design: In oled-col+ mode, cursor auto advances to next col (horizontally), in row+ mode, to the next row (vertically)

\ Note: I suspect a 128x64 OLED has a 8x 8-pixel high pages, whereas a 128x32 OLED has 8x 4-pixel high pages...



\ EXPLORATORY DESIGN
\ basic exported functions from 3iI2C.avr -- these use Wire.h for I2C comms
$1F constant CALL_I2C_INIT      \ start wire
$20 constant CALL_I2C_SETCLOCK  \ set clock speed
$21 constant CALL_I2C_OLED1306  \ send data transmission

\ register location to store I2C OLED display address - used by I2C_OLED1306
31 constant OLED_I2C_REGISTER
$3C constant OLED_I2C_ADDRESS  \ default address for SSD1306 based cheap OLEDs or $3D if set SMT resistor on back of board
\ TODO: what about read/write

\ SSD1306 commands 
$0  constant  SSD1306_COMMAND_FLAG
$40 constant  SSD1306_DATA_FLAG

\ Design: set which display size you're using
7 constant HEIGHT_128x64
3 constant HEIGHT_128x32
HEIGHT_128x64  constant SCREEN_HEIGHT


: i2c-init ( -- )   \ Wire begin - start I2C
    CALL_I2C_INIT xcall ;

: i2c-setclock ( val/10k -- ) \ e.g. 10 means 100kHz bitrate   \  TODO: see if faster clock (100, 800) means snappier display
    xstore CALL_I2C_SETCLOCK xcall ;

: i2c-restoreclock ( val/10k -- ) \ e.g. 10 means 100kHz bitrate
    xstore CALL_I2C_SETCLOCK xcall ;
\ TODO: CODE REVIEW - what timing should be restored to?

: oled-setreg ( i2c-adr -- ) \ store I2C write address into fixed register
            \ ex: $3C OLED-display with SSD1306 driver IC - note: address changes when switching between write/read
   OLED_I2C_REGISTER setreg  ;

\ CODE REVIEW: consider to use nstore instead of xstore for communication efficiency!       
: [oled-bytes]  ( b1 b2 b3 ... bn N -- )  \ sends N bytes in reverse order and puts N at the end (target TOS)
    >r r@ revstore 
    r> xstore ;
: [oled-send]   >r [oled-bytes] r> xstore CALL_I2C_OLED1306 xcall ;
\ Note: N bytes
: oled-cmd  ( c1 c2 ... cn N -- )  SSD1306_COMMAND_FLAG  [oled-send] ;
: oled-data ( d1 d2 ... dn N -- )  SSD1306_DATA_FLAG     [oled-send] ;

\ ===================================================
\ SSD1306 OLED driver IC commands -- see datasheet

\ contrast control 256-steps ($FF brightest, $0 dimmest), brighter consumes more current, $8F default
$81  constant  SSD1306_SETCONTRASTCONTROL
: oled-brightness ( val -- )  SSD1306_SETCONTRASTCONTROL  swap  2 oled-cmd ; 

\ NOTE: this command is not super useful.  would be better to lock display with all pixels off
$A4  constant  SSD1306_DISPLAYRAM
$A5  constant  SSD1306_DISPLAYON
: oled-display-block   SSD1306_DISPLAYON  1 oled-cmd ;  \ show all pixels active / GDDRAM can be changed in background / see unblock
: oled-display-unblock SSD1306_DISPLAYRAM 1 oled-cmd ;  \ show whatever's in GDDRAM

\ display mode
$A6  constant  SSD1306_NORMALDISPLAY   \ default, 1 turns pixel on, 0 turns pixel off
$A7  constant  SSD1306_INVERTDISPLAY   \ inverted logic (active low): 0 turns pixel on, etc.
: oled-normal  SSD1306_NORMALDISPLAY 1 oled-cmd ;
: oled-invert  SSD1306_INVERTDISPLAY 1 oled-cmd ;

\ display on/off
$AE  constant  SSD1306_DISPLAYOFF      \ default, sleep, data not affected
$AF  constant  SSD1306_DISPLAYON
: oled-sleep SSD1306_DISPLAYOFF 1 oled-cmd ;  \ blanks screen (does not affect image stored in GDDRAM)
: oled-wake  SSD1306_DISPLAYON  1 oled-cmd ;  \ shows screen

\ charge pump
$8D  constant  SSD1306_CHARGEPUMP
: oled-dim    SSD1306_CHARGEPUMP $10 2 oled-cmd ; \ disable (default)
: oled-bright SSD1306_CHARGEPUMP $14 2 oled-cmd ; \ enable


\ memory mode functions
$20  constant  SSD1306_MEMORYMODE      
  $00  constant  SSD1306_MEMORY_COL+
  $01  constant  SSD1306_MEMORY_ROW+
  $10  constant  SSD1306_MEMORY_PAGE+
$C0  constant  SSD1306_COMSCANINC      \ rowscan direction (commons remap)
$C8  constant  SSD1306_COMSCANDEC      
  $A0  constant  SSD1306_SEGSCANINC     \ colscan direction (segments remap)
  $A1  constant  SSD1306_SEGSCANDEC

\ memory address modes - first two are text mode, last is graphics mode
: oled-col+   SSD1306_MEMORYMODE  SSD1306_MEMORY_COL+ 2 oled-cmd ;  \ auto-advance to right then down, re-start at top left (horizontal increment)
: oled-page+  SSD1306_MEMORYMODE  SSD1306_MEMORY_PAGE+ 2 oled-cmd ;  \ auto-advance to right, but loop in same page - default
: oled-row+   SSD1306_MEMORYMODE  SSD1306_MEMORY_ROW+ 2 oled-cmd ;  \ auto-advance down, then right (vertical increment)

\ screen orientation (scanning setup) - allows installing display in any layout.   screen flips on the fly
: oled-pins-top       SSD1306_COMSCANDEC  SSD1306_SEGSCANDEC  2 oled-cmd   oled-col+ ;  
: oled-pins-bottom    SSD1306_COMSCANINC  SSD1306_SEGSCANINC  2 oled-cmd   oled-col+ ; 
: oled-pins-left      SSD1306_COMSCANDEC  SSD1306_SEGSCANINC  2 oled-cmd   oled-row+ ;
: oled-pins-right     SSD1306_COMSCANINC  SSD1306_SEGSCANDEC  2 oled-cmd   oled-row+ ;
\ TODO: in left/right mode, need the pixel expansion, instead of row expansion (only 8 rows)
\ TODO: add dynamic flip of display and correction of image on screen using an orientation sensor

\ scanning mode functions
$22  constant  SSD1306_ROWRANGE       \ row/page/common addr  Set Page Address
$21  constant  SSD1306_COLRANGE       \ col/segment     addr
$10  constant  SSD1306_SETHICOL       \ page addressing mode
$00  constant  SSD1306_SETLOCOL       \ page addressing mode
$B0  constant  SSD1306_SETROW         \ page start address

\ window navigation
: oled-window-full ( -- ) \ full screen, init cursor to top left.  restricts pixel changes to window range: row=[0,7], col= [0,$7F]
   SSD1306_ROWRANGE 0 SCREEN_HEIGHT    SSD1306_COLRANGE 0 $7F  6 oled-cmd ;  
: oled-rowstart    ( -- )   oled-page+  $10 $00 2 oled-cmd oled-col+ ; \ Reset col to 0 at start of current row
: oled-colselect   ( c -- ) oled-page+  SSD1306_SETHICOL /mod  
          SSD1306_SETHICOL or   swap    SSD1306_SETLOCOL or   2 oled-cmd oled-col+ ;  \ selects column
: oled-rowselect   ( r -- p ) 8 /mod    SSD1306_SETROW   or   1 oled-cmd ;  \ calculates which row, sets page, returns pixel in page


\ PIXEL GRAPHICS - NOTE: clobbers other pixels in same page: 8x1 (down x over) block of pixels
\ TODO: FIND IF THERE IS A WAY NOT TO CLOBBER OTHER PIXELS IN THE SAME PAGE, in order to successively add points without maintaining
\ a software buffer
\ TODO - read colbyte GDDRAM memory so can mask the additional pixel on/off or not (vs. create/save display buffer)
\ ONCE THIS IS DONE - then can finish proper etch-a-sketch... with this driver, and on ATTINY...

\ Design: p specifies pixel position
: set-pixel ( p -- ) 8 mod 1 swap lshift 1 oled-data ;   \ [0-7] pth pixel from top of colbyte, does wrap around 
: next-page ( -- ) oled-row+ $0 1 oled-data  oled-col+ ; \ next character row (block of 8 pixels)  \ sets left/right/down mode
: .cr ( -- ) next-page oled-rowstart ;

\ Set consecutive pixels
: set8   ( p -- ) 8 0 do dup set-pixel loop drop ;
: set64  ( p -- ) 8 0 do dup set8 loop drop ;
: set128 ( p -- ) 2 0 do dup set64 loop drop ;
: diag8  (   -- ) 8 0 do i set-pixel loop ;
: diag64 (   -- ) 8 0 do diag8 next-page loop ;


\ Positioning: r=row, c=column
: set-pos  ( r c -- p ) oled-colselect oled-rowselect ;  \ r in [0,63], c=[0,127],  p = pixel in the right page to hit row

: draw-dot ( r c -- )  set-pos set-pixel ;  \ row, col screen coordinates [0-63]x[0-127]

: clear-row ( r c -- ) dup -rot set-pos drop 128 swap do 0 1 oled-data loop ;
: clear-line ( r c -- ) dup -rot set-pos 128 rot do dup 1 oled-data loop drop ; \ TODO: Can't do this properly without READ

\ TODO: draw Turtle
\ TODO: Pacman
\ TODO: animation, snowflakes falling, drawing fractals, screensavers, or any graphical element...


variable page32
: page32++ page32 @ 16 + 64 mod page32 ! ;
: reset-page32 0 page32 ! ;


\ display primitives
16 constant MAX_I2C_PACKET  \ 32 is sometimes glitchy on Atmega328P  \ TODO: increase speed of I2C xfer to get visual snappyness
: [oled-fillpacket] ( asc -- asc )  MAX_I2C_PACKET 0    do dup loop    MAX_I2C_PACKET oled-data ;   
: [oled-fillrow]    ( asc -- asc )  128 MAX_I2C_PACKET / 0 do [oled-fillpacket] loop ;
\ FEATURE: store pins top/bottom state, and clear screen by flipping memory mode row+/col+ to get a nice smooth scrool
\ may not be as relevant when clearing directly from uC (faster)
: [oled-fill]  ( asc -- )  oled-window-full oled-row+ 8 0 do  [oled-fillrow]  loop  oled-col+ drop ;  \ sets left/right/down mode
: oled-clear-slow  $00 [oled-fill]  reset-page32 ;  \ slow-function
: oled-fill-slow   $FF [oled-fill] ;  \ slow-function




\ user functions
: oled-init-slow  \ sending INIT sequence from host - so can experiment with variations
  i2c-init
  1 i2c-setclock
  OLED_I2C_ADDRESS oled-setreg
  oled-bright
  $8F oled-brightness
  oled-normal
  
  oled-pins-top     
  oled-window-full
  oled-fill-slow
  oled-wake
  250 ms
  oled-clear-slow  \ sets left/right/down mode
;

\ ========================================================
\ TODO: Font experiments
\   8x6 instead of 8x8
\   Half size fonts
\   Double fonts
\   Italics
\   Quad fonts
\   128x32 displays

\ ========================================================

\ ADVANCED EXPORTED FUNCTIONS FROM 3II2C.AVR
$22 constant CALL_I2C_OLEDFILL
$23 constant CALL_I2C_OLEDIMAGE
$24 constant CALL_I2C_OLEDINIT
$25 constant CALL_I2C_OLEDCHAR

: oled-clear  $00 xstore CALL_I2C_OLEDFILL xcall ;   \ fast-function  -- no need for 500ms delay with I2C speed of 10,000
: oled-fill   $FF xstore CALL_I2C_OLEDFILL xcall ;   \ fast-function

: oled-image ( index -- )  xstore CALL_I2C_OLEDIMAGE xcall 500 ms ; \ fast-function, reads from flash memory

: oled-char   ( ascii -- )  xstore CALL_I2C_OLEDCHAR xcall ;
: oled-string ( sptr len -- ) 0 do c@+ oled-char loop drop ;  \  print string  Ex: s" hello" oled-string.  Much faster than ..oled
: .oled  oled-char ;
: ..oled oled-string ;


: oled-init  \ using INIT sequence embedded into target
  i2c-init
  1 i2c-setclock                    
  OLED_I2C_ADDRESS  oled-setreg
  CALL_I2C_OLEDINIT xcall
  oled-fill
  250 ms
  oled-clear
;


\ What's this?
: oled-init1 ( -- )
    $D5 $80  \ osc. frequency controlling frame frequency
    $A8 31   \ set rows (multiplex)
    $DA $12  \ set COM pins configuration
    $D9 $F1  \ set per-charge period
    8 oled-cmd ;
    
: oled-init2 ( -- )
    $D3 $0    \ start row (offset)
    $40 $0 or \ memory map start row
    3 oled-cmd ;

: oled-init6 ( -- )
    $DB $40 $A4 $A6 $2E $AF  6 oled-cmd ;


\ ================= Smaller displays

\ Note: 
\ 0 0 draw-dot  \ even rows write invisibly onto the page, clobbering the page, but leaving no mark of their own
\ 1 0 draw-dot  \  odd rows are active
\ Hypothesis --- the 128x32 displays only display on the odd rows, hence we need a new way to map pages, combining two pages...





\ ================= FONTS

( DISCOVERY Forth for Fonts )
( This is the fixed-width bitmap font from glcdfont.c for Adafruit_GFX )
( Converted to GForth by Assad Ebrahim - 2021-05-29---Sat---16:06 )
( 3iI2C.fs has fast character display with embedded font management using 3iFonts.avr -- 2021-07-24---Sat---02:41 )
( This file is now for the Forth driven slower code working from scratch  - no font info at all)
 

\ ASCII 5x7 wxh pixel font, 5 bytes each, so 1.3kB of memory for the full ASCII 256 characters, and 0.7kB for 128 chars
\ these are rotated cw 90 degrees so need to unrotate them.  
\ Why store this way?  to store them unrotated, would take more memory, 7 bytes using only 5 bits each, so X% less efficient
\ Also ** text-oriented OLED/LCD devices with text-oriented device driver ICs such as SSD1306 want the bytes in vertical format
\ Note: When sending to an OLED display device in bulk, we reverse them loading into the uC so hence the read from rawfont is fwd.

create rawfont  
	$00 c, $00 c, $00 c, $00 c, $00 c,
	$3E c, $5B c, $4F c, $5B c, $3E c,
	$3E c, $6B c, $4F c, $6B c, $3E c,
	$1C c, $3E c, $7C c, $3E c, $1C c,
	$18 c, $3C c, $7E c, $3C c, $18 c,
	$1C c, $57 c, $7D c, $57 c, $1C c,
	$1C c, $5E c, $7F c, $5E c, $1C c,
	$00 c, $18 c, $3C c, $18 c, $00 c,
	$FF c, $E7 c, $C3 c, $E7 c, $FF c,
	$00 c, $18 c, $24 c, $18 c, $00 c,
	$FF c, $E7 c, $DB c, $E7 c, $FF c,
	$30 c, $48 c, $3A c, $06 c, $0E c,   \ TEST: first non-symmetric row is:     11 showfont
	$26 c, $29 c, $79 c, $29 c, $26 c,
	$40 c, $7F c, $05 c, $05 c, $07 c,
	$40 c, $7F c, $05 c, $25 c, $3F c,
	$5A c, $3C c, $E7 c, $3C c, $5A c,
	$7F c, $3E c, $1C c, $1C c, $08 c,
	$08 c, $1C c, $1C c, $3E c, $7F c,
	$14 c, $22 c, $7F c, $22 c, $14 c,
	$5F c, $5F c, $00 c, $5F c, $5F c,
	$06 c, $09 c, $7F c, $01 c, $7F c,
	$00 c, $66 c, $89 c, $95 c, $6A c,
	$60 c, $60 c, $60 c, $60 c, $60 c,
	$94 c, $A2 c, $FF c, $A2 c, $94 c,
	$08 c, $04 c, $7E c, $04 c, $08 c,
	$10 c, $20 c, $7E c, $20 c, $10 c,
	$08 c, $08 c, $2A c, $1C c, $08 c,
	$08 c, $1C c, $2A c, $08 c, $08 c,
	$1E c, $10 c, $10 c, $10 c, $10 c,
	$0C c, $1E c, $0C c, $1E c, $0C c,
	$30 c, $38 c, $3E c, $38 c, $30 c,
	$06 c, $0E c, $3E c, $0E c, $06 c,
	$00 c, $00 c, $00 c, $00 c, $00 c,
	$00 c, $00 c, $5F c, $00 c, $00 c,
	$00 c, $07 c, $00 c, $07 c, $00 c,
	$14 c, $7F c, $14 c, $7F c, $14 c,
	$24 c, $2A c, $7F c, $2A c, $12 c,
	$23 c, $13 c, $08 c, $64 c, $62 c,
	$36 c, $49 c, $56 c, $20 c, $50 c,
	$00 c, $08 c, $07 c, $03 c, $00 c,
	$00 c, $1C c, $22 c, $41 c, $00 c,
	$00 c, $41 c, $22 c, $1C c, $00 c,
	$2A c, $1C c, $7F c, $1C c, $2A c,
	$08 c, $08 c, $3E c, $08 c, $08 c,
	$00 c, $80 c, $70 c, $30 c, $00 c,
	$08 c, $08 c, $08 c, $08 c, $08 c,
	$00 c, $00 c, $60 c, $60 c, $00 c,
	$20 c, $10 c, $08 c, $04 c, $02 c,
	$3E c, $51 c, $49 c, $45 c, $3E c,
	$00 c, $42 c, $7F c, $40 c, $00 c,
	$72 c, $49 c, $49 c, $49 c, $46 c,
	$21 c, $41 c, $49 c, $4D c, $33 c,
	$18 c, $14 c, $12 c, $7F c, $10 c,
	$27 c, $45 c, $45 c, $45 c, $39 c,
	$3C c, $4A c, $49 c, $49 c, $31 c,
	$41 c, $21 c, $11 c, $09 c, $07 c,
	$36 c, $49 c, $49 c, $49 c, $36 c,
	$46 c, $49 c, $49 c, $29 c, $1E c,
	$00 c, $00 c, $14 c, $00 c, $00 c,
	$00 c, $40 c, $34 c, $00 c, $00 c,
	$00 c, $08 c, $14 c, $22 c, $41 c,
	$14 c, $14 c, $14 c, $14 c, $14 c,
	$00 c, $41 c, $22 c, $14 c, $08 c,
	$02 c, $01 c, $59 c, $09 c, $06 c,
	$3E c, $41 c, $5D c, $59 c, $4E c,
	$7C c, $12 c, $11 c, $12 c, $7C c,
	$7F c, $49 c, $49 c, $49 c, $36 c,
	$3E c, $41 c, $41 c, $41 c, $22 c,
	$7F c, $41 c, $41 c, $41 c, $3E c,
	$7F c, $49 c, $49 c, $49 c, $41 c,
	$7F c, $09 c, $09 c, $09 c, $01 c,
	$3E c, $41 c, $41 c, $51 c, $73 c,
	$7F c, $08 c, $08 c, $08 c, $7F c,
	$00 c, $41 c, $7F c, $41 c, $00 c,
	$20 c, $40 c, $41 c, $3F c, $01 c,
	$7F c, $08 c, $14 c, $22 c, $41 c,
	$7F c, $40 c, $40 c, $40 c, $40 c,
	$7F c, $02 c, $1C c, $02 c, $7F c,
	$7F c, $04 c, $08 c, $10 c, $7F c,
	$3E c, $41 c, $41 c, $41 c, $3E c,
	$7F c, $09 c, $09 c, $09 c, $06 c,
	$3E c, $41 c, $51 c, $21 c, $5E c,
	$7F c, $09 c, $19 c, $29 c, $46 c,
	$26 c, $49 c, $49 c, $49 c, $32 c,
	$03 c, $01 c, $7F c, $01 c, $03 c,
	$3F c, $40 c, $40 c, $40 c, $3F c,
	$1F c, $20 c, $40 c, $20 c, $1F c,
	$3F c, $40 c, $38 c, $40 c, $3F c,
	$63 c, $14 c, $08 c, $14 c, $63 c,
	$03 c, $04 c, $78 c, $04 c, $03 c,
	$61 c, $59 c, $49 c, $4D c, $43 c,
	$00 c, $7F c, $41 c, $41 c, $41 c,
	$02 c, $04 c, $08 c, $10 c, $20 c,
	$00 c, $41 c, $41 c, $41 c, $7F c,
	$04 c, $02 c, $01 c, $02 c, $04 c,
	$40 c, $40 c, $40 c, $40 c, $40 c,
	$00 c, $03 c, $07 c, $08 c, $00 c,
	$20 c, $54 c, $54 c, $78 c, $40 c,
	$7F c, $28 c, $44 c, $44 c, $38 c,
	$38 c, $44 c, $44 c, $44 c, $28 c,
	$38 c, $44 c, $44 c, $28 c, $7F c,
	$38 c, $54 c, $54 c, $54 c, $18 c,
	$00 c, $08 c, $7E c, $09 c, $02 c,
	$18 c, $A4 c, $A4 c, $9C c, $78 c,
	$7F c, $08 c, $04 c, $04 c, $78 c,
	$00 c, $44 c, $7D c, $40 c, $00 c,
	$20 c, $40 c, $40 c, $3D c, $00 c,
	$7F c, $10 c, $28 c, $44 c, $00 c,
	$00 c, $41 c, $7F c, $40 c, $00 c,
	$7C c, $04 c, $78 c, $04 c, $78 c,
	$7C c, $08 c, $04 c, $04 c, $78 c,
	$38 c, $44 c, $44 c, $44 c, $38 c,
	$FC c, $18 c, $24 c, $24 c, $18 c,
	$18 c, $24 c, $24 c, $18 c, $FC c,
	$7C c, $08 c, $04 c, $04 c, $08 c,
	$48 c, $54 c, $54 c, $54 c, $24 c,
	$04 c, $04 c, $3F c, $44 c, $24 c,
	$3C c, $40 c, $40 c, $20 c, $7C c,
	$1C c, $20 c, $40 c, $20 c, $1C c,
	$3C c, $40 c, $30 c, $40 c, $3C c,
	$44 c, $28 c, $10 c, $28 c, $44 c,
	$4C c, $90 c, $90 c, $90 c, $7C c,
	$44 c, $64 c, $54 c, $4C c, $44 c,
	$00 c, $08 c, $36 c, $41 c, $00 c,
	$00 c, $00 c, $77 c, $00 c, $00 c,
	$00 c, $41 c, $36 c, $08 c, $00 c,
	$02 c, $01 c, $02 c, $04 c, $02 c,
	$3C c, $26 c, $23 c, $26 c, $3C c,
	$1E c, $A1 c, $A1 c, $61 c, $12 c,
	$3A c, $40 c, $40 c, $20 c, $7A c,
	$38 c, $54 c, $54 c, $55 c, $59 c,
	$21 c, $55 c, $55 c, $79 c, $41 c,

: getfont-fwd ( n -- b1 b2 b3 b4 b5 )   \ reads in order (left to right), indexed from 0  GOOD FOR OLED OUTPUT
    5 * rawfont + 5 0 do c@+ swap loop drop ;
: getfont-rev ( n -- b5 b4 b3 b2 b1 )   \ reads in reverse into stack so display is easier  GOOD FOR LAPTOP OUTPUT WITH ROTATION
    1 + 5 * 1 - rawfont + 5 0 do c@- swap loop drop ;  
: getfont ( n -- b5 b4 b3 b2 b1 )  \ SET TO WHAT YOU WANT IT TO BE FOR YOUR APP
    getfont-rev ;

\ Display directly to screen
42 constant star
: .font ( b -- ) 0> if star else bl then emit ;
: font-line ( b -- )   \ display one line of a 5-byte font
   T8 init-mask begin 2dup and .font 1 rshift dup 0= until 2drop ;
: [showfont] ( b1 ... b5 -- ) 5 0 do cr font-line loop cr ;
: showfont ( n -- ) dup emit getfont-rev [showfont] ;

\ Display to screen: rotated ccw 90 deg, use * and bl for pixels
: [rotfont] ( b1 b2 b3 b4 b5 )  \ walks through and rasters by columns, outputting pixel at a time the row orientation
    7 0 do cr \ j = outside loop
    5 0 do 4 roll dup 1 j lshift and .font loop 
    loop cr 5 ndrop ;
: rotfont ( n -- ) dup emit getfont-fwd [rotfont] ;
: demofont ( -- ) 58 48 do cr i rotfont loop ;  \ show 0-9 to PC screen


\ FONT OUTPUT TO OLED I2C with SSD1306 driver chip (oled-init)
\ on a 128x64 OLED, this means 16x8 characters, or 128 characters total
: .bl ( -- ) 0 0 0 0  0 0 0 0  8 oled-data ;  \ space character (16 per row on a 128x64 pixel display)
: .lf ( -- ) oled-row+ $0 1 oled-data oled-col+ ; \ linefeed: in row+ mode, sending $00 moves ptr to next OLED page (row)
: .cr ( -- ) .lf oled-rowstart ; 

: pad ( p# -- p# ) dup >r 0 do 0 loop r> ;  \ fill out 3 pixels after character emitted (makes it an 8x8 square)
: .oled ( ascii -- ) getfont-fwd  1 pad  5 + oled-data ; \ emit character with 3 pixels spacing to oled
: ..oled ( sptr len -- ) 0 do c@+ .oled loop drop ;  \  print string  Ex: s" hello" ..oled
: ..oled-diag ( sptr len -- ) 0 do c@+ .oled .lf loop drop ; \ diagonal print


\ TEST routines
: oled-ascii ( -- ) oled-clear 128 0 do i dup .oled 10 mod 0 > if else key drop then loop ; 
: oled-hello ( -- ) s" Hello World!" ..oled ;

: oled-num8  s" 12345678" ..oled ;
: oled-num10 s" 0123456789" ..oled ;
: oled-numrow oled-num8 oled-num8 ;
: oled-num8fill 8 0 do oled-numrow loop ;
: oled-num8fill 8 0 do oled-numrow loop ;
: oled-num8fill 8 0 do oled-numrow loop ;
: oled-num10fill 13 0 do oled-num10 loop ;

\ can fill 16x8 char (total 128) in a 128x64 pixel screen  (each character is 8x8 pixel block)
variable count 
: count! ( n -- ) count c! ;
: count@ ( -- n ) count c@ ;
: count++ ( -- c' ) count@ 1 + 10 mod dup count! ;
\ Fills screen with continuously increasing text --- good for testing scrolling functionality
: oled-textroll count@ 128 0 do '0' + .oled count++ loop drop ;


: demo-flip
  \ pins top
  oled-pins-top  oled-window-full oled-clear 
  s" Hello World!" ..oled
  key drop
  \ pins down
  oled-pins-bottom oled-window-full oled-clear 
  s" Hello World!" ..oled
;

\ TODO: add transposed character for writing in row+ mode
\ TODO: DEMONSTRATE writing in row+ mode,  unrotated and rotated


\ ---- WRITING to 32 high oled displays ----------------------
\ Notes: Reverse engineering using 128x32 pixel oled display
\ Notes: It appears the 32 high display still has 8 rows of 4.
\ Notes: Hardware wise, it still has 64 COM (rows), but it appears to map the first (0) and every other even COM to ground so that the SEG (col) would be cleared if any of those are triggered --- prove with draw-dot
\    0 0 draw-dot  --> clear
\    1 0 draw-dot
\    2 0 draw-dot  --> clear
\    3 0 draw-dot

\ Then 1010 1010b should draw 4 consecutive vertical dots its page of 4-high
\     $AA 1 oled-data
\ How to get a page of 8?  Two successive 4-high pages...
\     oled-row+
\     $AA $AA 2 oled-data
\ How to write just an 8-high page and then go back to next col?
\     oled-row+ $AA 1 oled-data  ok
\     oled-col+ $AA 1 oled-data  ok
\     0 oled-rowselect drop

\ How to expand a byte across 2 bytes?
\ used for generating 8-high char for 32-high oled display
: xf32  ( ascii  --  )  \ loop through byte from MSB side, insert 0 bit, resulting in 2 bytes - 
    T8 init-mask begin 
        2dup and bit >r 0 >r 
        1 rshift  \ advance mask
        dup 0=    \ test if done
    until 2drop 
    r> r> r> r> r> r> r> r> 
    r> r> r> r> r> r> r> r> 
    16 nrev drop 
    >b8 >r >b8 r> ;

: .32 ( ascii -- ) xf32 
    oled-row+ 1 oled-data 
    oled-col+ 1 oled-data
    page32 @ oled-rowselect drop
;

: .oled32 ( ascii -- ) getfont-rev 1 pad 5 + 0 do .32 loop ;
: ..oled32 ( sptr len -- ) 0 do c@+ .oled32 loop drop ;  \  print string  Ex: s" hello" ..oled
: .cr32 oled-row+ $0 $0 2 oled-data oled-rowstart page32++ oled-col+ ;

: demowrite32 
    oled-clear-slow
    s" Hello world!" ..oled32 
    .cr32 s" Goodbye world!" ..oled32
    .cr32 s" ... till we meet again2" ..oled32
    .cr32 s" in the hereafter." ..oled32
    .cr32 s" Rewriting history." ..oled32
;

\ TODO: build this in for fast display!


\ ------------- MENUS --------------------

: demo-menu \ works well with yellow-blue OLED screens 128x64
oled-clear
0 10 set-pos s" MENU" ..oled
16 20 set-pos s" A. Choice 1" ..oled
24 20 set-pos s" B. Choice 2" ..oled
32 20 set-pos s" C. Choice 3" ..oled
40 20 set-pos s" D. Choice 4" ..oled
56 10 set-pos s" Select option" ..oled
;

: menu-breakfast \ works well with yellow-blue OLED screens 128x64
oled-clear
0 10 set-pos s" BREAKFAST" ..oled
16 20 set-pos s" A. Coco Pops" ..oled
24 20 set-pos s" B. Rice Krispy" ..oled
32 20 set-pos s" C. Egg McMuffy" ..oled
40 20 set-pos s" D. Alpha Bites" ..oled
56 10 set-pos s" Select option" ..oled
;


\ TODO: Port fonts to ATtiny85...
\ TODO: Create MenuSystem on ATtiny85... with left arrow and 4 buttons that engage the presses...
\ TODO: Game: guess my number
\ TODO: Game: tone and sequence guesser


\ ==========================

: help-font
   cr
   cr ." 3iFont.fs (help-font)"
   cr ." ====================="
   cr
   cr ." include 3iForth.fs  3iI2C.fs  3iFont.fs"
   cr 
   cr ." 1 open-com   rx"
   cr ." oled-init"
   cr 
   cr ." Low-Level"
   cr ." ---------"
   cr ." (on PC screen)
   cr ." 48 getfont  font-line"  
   cr ." 48 showfont 48 rotfont"
   cr ." 48  42  65  'a'"
   cr ." demofont "
   cr
   cr ." High-Level"
   cr ." ----------"
   cr ." (on oled-display)"
   cr ." char:  48  'a'  .oled"
   cr ." string:  <squote> Hello World!<squote> ..oled"
   cr ." demo-menu"   
;

help-font


\ VERSION CONTROL
\ V2 - 2021-06-10---Thu---23:45 - Added OLED out capability via I2C only (no software driver) 
\ V3 - 2021-07-24---Sat---02:42 - ADDED embedded font and fast character display 


\ Next steps:
\ + Have 1 to 4 pixel vertical spacing between rows of text
\ + Position under software control rather than the auto advance of the screen
\ + Pixel display
\ + I2C under Attiny
\ + Then can embed displays with Attiny :)
\ + Including short, tight menus
\ + EXERCISE (fun with Jasmine) - create own font! :)  Cursive handwriting program...  font editor?
\    https://en.wikipedia.org/wiki/Computer_font
\ + What about musical notes? on a staff -- can create own font for this...
\ + Scaling fonts - how to enlarge so we get bigger characters?
\ + Font types - bitmap is a matrix of pixels
\ + Bitmap graphics?  Bitmap file format?
\    https://en.wikipedia.org/wiki/BMP_file_format
\ + EXERCISE: Get AMPL logo (jaguar head) to show
\ + EXERCISE: Display a small photo... monochrome
\ + EXERCISE: ASCII Art
\    https://www.dreamstime.com/royalty-free-stock-images-typed-obama-image11549279
\ Very nice conversion tool:
\ https://www.dcode.fr/binary-image

\ 48x48 pixel AMPL jaguar outline, black text on white background
\ see 3iArt.fs

\ EXERCISE: crossword puzzle / word search
\ EXERCISE: writing column orientation by flipping the font

\ TODO: 2021-07-24---Sat---13:38 - do 2x scaling, 3x scaling for fonts...



\ =================  APPLICATIONS


\ ==== CHARACTERS ====  See 3iFonts.fs

: .note   14   .oled ;
: .rest   22   .oled ;
: .space  32   .oled ;
: .minus  45   .oled ; 
: .rarrow 16   .oled ;
: .oarrow 111  .oled ;
: .larrow 17   .oled ;
: .plus   43   .oled ; 
: .cross  88   .oled ;
: .star   42   .oled ;

\ =================  SOFTWARE MENUS: 
\ - < o > +
\ ESC < OK > +
\ BACK < OK > +
\ X < TICK > +
: .menu5  oled-init  .minus .space .larrow .space .oarrow .space .rarrow .space .plus .cr  ;

: menu-fast \ works well with yellow-blue OLED screens 128x64
oled-clear
0 10 set-pos s" BREAKFAST" oled-string
16 20 set-pos s" A. Coco Pops" oled-string
24 20 set-pos s" B. Rice Krispies" oled-string
32 20 set-pos s" C. Egg McMuffin" oled-string
40 20 set-pos s" D. Alpha Bites" oled-string
56 10 set-pos s" Select option" oled-string
;

\ ================ SOFTWARE DICE:
\ need random number roller
\ APPLICATION: D&D / Fighting Fantasy accompaniment...
: .dice6  oled-clear .cr .cr          s" * * *"     ..oled  .cr .cr                  .cr .cr s" * * *" ..oled                      ;
: .dice5  oled-clear .cr .cr          s" *   *"     ..oled  .cr .cr s"   *  " ..oled .cr .cr s" *   *" ..oled ;
: .dice4  oled-clear .cr .cr          s" *   *"     ..oled  .cr .cr                  .cr .cr s" *   *" ..oled                      ;
: .dice3  oled-clear .cr .cr          s" *"         ..oled  .cr .cr s"   *"   ..oled .cr .cr s"     *" ..oled ;
: .dice2  oled-clear .cr .cr                                .cr .cr s" *   *" ..oled .cr .cr ;
: .dice1  oled-clear .cr .cr                                .cr .cr s"   *"   ..oled .cr .cr ;
create .dice ' .dice1 , ' .dice2 , ' .dice3 , ' .dice4 , ' .dice5 , ' .dice6 ,
: dice-show ( n -- ) 1 - cells .dice + @ execute ;

\ ====================================
\ GRAPHICS

\ EXERCISE - Animated square moving across and down the screen
\ TODO: keys input -- controlling the ball
\ TODO: random perturbations --- computer trying too swing the sphere, you need to control it, timer --- mobius the game :)

: top>b8 ( b1 b2 b3 b4 b5 b6 b7 b8 -- vb ) \ produce vertical byte with b1 as lsb
    8 nrev drop >b8 ;

: 4x4sqA ( r c -- ) 
  set-pos drop
  $00 3 0 do dup loop 
  $0F 3 0 do dup loop 
  8 oled-data ;

: 4x4sqB ( r c -- ) 
  set-pos drop
  $0F 3 0 do dup loop 
  $00 3 0 do dup loop 
   8 oled-data ;
  
: 4x4sqC ( r c -- )
  set-pos drop
  $F0 3 0 do dup loop 
  $00 3 0 do dup loop 
   8 oled-data ;

: 4x4sqD ( r c -- )
  set-pos drop
  $00 3 0 do dup loop 
  $F0 3 0 do dup loop 
   8 oled-data ;
   
variable magic#
: magic! ( n -- ) magic# ! ;
: magic@ ( -- n ) magic# @ ;
: magic++ ( -- # ) magic@ 3 + dup magic! ;
: reset-magic ( r c -- r c ) 2dup swap 8 / 128 * + magic# ! ;
: magic-pos ( r c -- r' c' ) 2drop magic++ 128 /mod 8 * 64 mod swap ;

: 4x4sq ( r c -- )  \ animation
  oled-clear
  reset-magic
  begin 
    2dup 4x4sqA 10 ms
    2dup 4x4sqB 10 ms
    2dup 4x4sqC 10 ms
    2dup 4x4sqD 10 ms
    magic-pos     \ slow drift along columns and rows
\    cr .s key drop
  key?
  until
;

\ TODO: game with a tilt / IMU sensor where you angle it and the "ball" moves across the screen and you need to guide it through 
\ a maze and into a pocket...

\ Graphic - tiles, 6x6 is a grid
: 8x8sq
  $FF 7 0 do dup loop 8 oled-data ;
: 6x6sq
  $0 $7E 5 0 do dup loop $0 8 oled-data ;
: tile oled-clear 128 0 do 6x6sq loop ;


\ ================= DRAWING PROGRAM

\ Design: getkey encodes escaped characters using GForth ekey and maps them to 0-9
\ Design: ENTER is 13, ESC is 27 (all above 9)
\ Example GForth keyboard control loop with drawing program   
: getkey ( -- k) ekey $FF and ;  \ k in {0,1,2,3}  In GForth, key does not give escaped characters properly, so use ekey
: lt ( 0 ." left"  )  1 -   128 mod         ; \ col-
: rt ( 1 ." right" )  1 +   128 mod         ; \ col+
: up ( 2 ." up"    ) swap 8 - 64 mod swap   ; \ row-
: dn ( 3 ." down"  ) swap 8 + 64 mod swap   ; \ row+
: hm ( 4 ." home"  ) 2drop 0 0              ; \ go home
: ed ( 5 ." end"   ) 2drop 56 120           ; \ go end
: pu ( 6 ) ." pen-up"  0 ; 
: pd ( 7 ) ." pen-down"  0 ; 
: bs ( 8 ) ." backspace"  0 ; 
: de ( 9 ) ." delete"  0 ;     
create response ' lt , ' rt , ' up , ' dn , ' hm , ' ed , ' pu , ' pd , ' bs , ' de ,
: say-esc ." esc" .s cr  oled-clear  ;

: move-ball ( r c -- r' c' 0/1 )  \ has to be fat 8x8sq brush to avoid the problem of clobbering pixels in pages...
    2dup set-pos drop 8x8sq 
    getkey dup 0 9 between? 
        if cells response + @ execute 0
        else esc-key = 
            if say-esc 0 else -1 then 
        then 
    >r .s r> cr ; 
: [oled-draw] ( r c -- r' c' )  oled-clear  cr ( ." start" ) .s cr begin move-ball until ;    
: oled-draw ( -- )  0 0 [oled-draw] 2drop ;
\ TODO: Add delete
\ TODO: Add pen-up
\ TODO: Add read GDDRAM and save for re-creation (as an image)
\ TODO: Move to fine brush, fat brush, i.e. 8x8 -> 4x4 -> 2x2 -> 1x1...
\ DONE - Add home and end


\ ===================  PHOTOBOOK / TINYPHOTO

: oled-photobook  
oled-clear
s" TINY PHOTO!" oled-string .cr
s" by Assad Ebrahim" oled-string .cr
s" www.mathscitech.org" oled-string .cr
s" /articles/TinyPhoto" oled-string .cr
3000 ms
begin 7 2 do oled-clear i oled-image 1500 ms loop key? until 
;


\ ====================================

: help-i2c
  cr
  cr ." Help I2C (help-i2c)"
  cr
  cr ." include   3iForth.fs    3iI2C.fs    [3iFont.fs]   3iArt.fs"
  cr
  cr ." 1 open-com   rx"
  cr
  cr ." oled-init"
  cr 
  cr ." Low level:"
  cr ." ---------"
  cr ." oled-sleep  oled-wake"
  cr ." oled-bright  oled-dim"
  cr ." [0-255] oled-brightness"
  cr ." oled-fill  oled-clear" 
  cr ." oled-normal  oled-invert"
  cr ." oled-pins-top  oled-pins-bottom"
  cr
  cr ." High level:"
  cr ." ----------"
  cr ." r c set-pos    r c draw-dot"
  cr ." 11 oled-char  'a' .oled"
  cr ." s'' hello'' oled-string   s'' hello'' ..oled"
  cr ." oled-draw  (interactive drawing program with arrows)"
  cr ." oled-photobook  (only if #define photo in 3iI2C)"
  cr
;

help-i2c


\ need I2C communication primitives (see Wire.h/.cpp)
\ primitive is Arduino's setClock() function which sets clock speed / crystal
\ safe default for I2C is 100 kHz (standard speed) or 400 kHz (fast).  Nano clock speed is 16MHz.
\ CPU frequency is scaled so it cannot generate all possible frequencies

\ : setclk
\    TWBR = ((F_CPU / frequency) - 16) / 2;  \ sets register  .e.g. 16000 / 100 ->72  - note this uses TWI (two-wire interface)

\ ==========================================

( Writing this by translating the driver software into Forth.  )
( There is so much to discard in C! C code feels like obfuscation-ware )

\ Version History:
\ V0 - 2021-06-03---Thu---14:24  - started
\    - 2021-06-08---Tue---13:13 - revisiting following completion of 3iI2C.avr
\ V1 - 2021-06-09---Wed---04:27  - WORKS! with revised 3iForth.fs comms protocol
\ V2 - 2021-07-20---Tue---01:25 - completed set-pos, draw-dot, and clear-row capabilities
\ V3 - 2021-07-21---Wed---06:08 - FIXED snappy display problem -- set I2C_SPEED to 10k
\ V4 - 2021-07-24---Sat---02:42 - ADDED embedded font and fast character display 
\ V5 - 2021-07-25---Sun---01:00 - ADDED moving 4x4 sq animation.  ADDED drawing program with 8x8 squares and GForth key presses...
\ V6 - 2021-11-12---Fri---04:57 - Fixed no exit key bug in drawing program!

\ Next Steps:
\   + Embedded turtle logo Forth!
\   + Joystick or key or button control with an OLED I2C display
\   + If use TFT screen can actually do it with a TFT display
\   + positioning cursor arbitrarily > row+ writing rotated/unrotated.  Pixel activation...
\   + lf and cr (newline)
\   + easy display scrolling

\   + Graphics: how to set the ROW?  
\   + How to read the RAM to mask (set pixel) without clobbering other pixels?
\       Maybe it's to set the address to the READ mode, i.e. 0x3D instead of 3C?
\   
\   + Scrolling screen...

\   REFLASH THE FIRMWARE - SIMPLIFIED THE I2C OLED INTERFACE BY REMOVING RESTORECLOCK CMD (forth.avr, i2c.avr, forth.fs, i2c.fs)

\   + Graphics - Piano visual display simulator
\   + Graphics - Pong the game
\   + Addressible LEDs: - https://www.thesmarthomehookup.com/the-complete-guide-to-selecting-individually-addressable-led-strips/


\ include 3iI2C.fs
\ include 3iFont.fs
\ include 3iI2C.fs
\ include 3iArt.fs
\ include 3iArt-butterfly.fs
\ include 3iArt-portraits.fs

